@@ -1,7 +1,7 @@
 from . import balancer
 from . import kineticizer
 from . import misc
-from . import parameter_balancing
+from . import parameter_balancing_core
 from . import SBtab
 from . import tablibIO
 from . import validatorSBtab
