@@ -1,4 +1,7 @@
 #!/usr/bin/python
+'''
+A collection of useful tools and functions for the manipulation of SBtab tables and SBtab documents.
+'''
 import re
 import string
 import libsbml
@@ -9,10 +12,10 @@ import scipy.optimize
 import random
 import copy
 import math
-try:
-    from . import SBtab
-except:
-    import SBtab
+import os
+
+try: from . import SBtab
+except: import SBtab
 
 
 def table_type(sbtab):
@@ -29,11 +32,40 @@ def table_type(sbtab):
     return False
 
 
+def count_tabs(sbtab_string):
+    '''
+    Counts how many SBtabs are in in a given string.
+
+    Parameters
+    ----------
+    sbtab_string: str
+        SBtab table or tables in string representation.
+        
+    Returns: int
+        Amount of SBtab tables in given string.
+    '''
+    counter = 0
+    for row in sbtab_string.split('\n'):
+        if row.startswith('!!SBtab') or row.startswith('!!ObjTables'):
+            counter += 1
+    return counter
+
+
 def validate_file_extension(file_name, file_type):
     '''
-    returns Boolean to evaluate if the file has the correct extension:
+    Returns Boolean flag to evaluate if the file has the correct extension:
     sbml => xml
-    sbtab => tsv
+    sbtab => tsv, csv, xlsx.
+
+    Parameters
+    ----------
+    file_name: str
+        Name of the file.
+    file_type: str
+        Type of the file ('sbtab' or 'sbml').
+
+    Returns: Bool
+        Boolean flag indicating if the given file extension corresponds to the given file type.
     '''
     # check extension for sbml file
     if file_type == 'sbml' and file_name[-3:] == 'xml': return True
@@ -42,6 +74,8 @@ def validate_file_extension(file_name, file_type):
 
     # check extension for sbtab file
     if file_type == 'sbtab' and file_name[-3:] == 'tsv': return True
+    elif file_type == 'sbtab' and file_name[-3:] == 'csv': return True
+    elif file_type == 'sbtab' and file_name[-4:] == 'xlsx': return True
     elif file_type == 'sbtab': return False
     else: pass
 
@@ -51,18 +85,28 @@ def validate_file_extension(file_name, file_type):
 
 def check_delimiter(sbtab_file):
     '''
-    determine the delimiter of the tabular file
+    Determines the delimiter of the SBtab table
+
+    Parameters
+    ----------
+    sbtab_file: str
+        SBtab table in string representation.
+
+    Returns: str
+        Delimiter of the SBtab table ('\t', ',', or ';')
     '''
     sep = False
 
-    for row in sbtab_file.split('\n'):
-        if row.startswith('!!'): continue
-        if row.startswith('!'):
-            s = re.search('(.)(!)', row[1:])
-            # if there is only 1 column, we have to define a default separator
-            # let's use a tab.
-            try: sep = s.group(1)
-            except: sep = '\t'
+    try:
+        for row in sbtab_file.split('\n'):
+            if row.startswith('!!'): continue
+            if row.startswith('!'):
+                s = re.search('(.)(!)', row[1:])
+                # if there is only 1 column, we have to define a default separator
+                # let's use a tab.
+                try: sep = s.group(1)
+                except: sep = '\t'
+    except: pass
 
     return sep
 
@@ -75,10 +119,10 @@ def valid_prior(sbtab_prior):
     validity = []
 
     # check table type
-    if sbtab_prior.table_type != 'Quantity':
+    if sbtab_prior.table_type != 'QuantityInfo':
         validity.append('Error: The TableType of the prior file is not '\
                         'correct: %s. '\
-                        'It should be Quantity' % sbtab_prior.table_type)
+                        'It should be QuantityInfo' % sbtab_prior.table_type)
 
     # check for required columns
     required_columns = ['!QuantityType', '!Unit', '!MathematicalType',
@@ -154,30 +198,45 @@ def readout_config(sbtab_options):
                        'overwrite_kinetics', 'cell_volume', 'parametrisation',
                        'enzyme_prefactor', 'default_inhibition',
                        'default_activation', 'model_name', 'boundary_values',
-                       'samples']
+                       'samples', 'size_limit']
 
 
-    if '!Option' not in sbtab_options.columns_dict:
-        log.append('Error: The crucial option column is missing from the'\
+    if '!ID' not in sbtab_options.columns_dict:
+        log.append('Error: The crucial option (ID) column is missing from the'\
                    'options file')
     if '!Value' not in sbtab_options.columns_dict:
             log.append('Error: The crucial value column is missing from the'\
                        'options file')
 
     for row in sbtab_options.value_rows:
-        if row[sbtab_options.columns_dict['!Option']] not in allowed_options:
+        if row[sbtab_options.columns_dict['!ID']] not in allowed_options:
             log.append('There was an irregular option in the options file:'\
-                       '%s' % row[sbtab_options.columns_dict['!Option']])
+                       '%s' % row[sbtab_options.columns_dict['!ID']])
         else:
             if row[sbtab_options.columns_dict['!Value']] == '':
                 log.append('There is no value set for option:'\
-                           '%s' % row[sbtab_options.columns_dict['!Option']])
+                           '%s' % row[sbtab_options.columns_dict['!ID']])
     
-        parameter_dict[row[sbtab_options.columns_dict['!Option']]] = row[sbtab_options.columns_dict['!Value']]
+        parameter_dict[row[sbtab_options.columns_dict['!ID']]] = row[sbtab_options.columns_dict['!Value']]
 
     return parameter_dict, log
 
 
+def size_warning(sbml_file):
+    '''
+    check whether a given sbml file has more than 250 reactions and 
+    then yield a warning message for the online interface
+    '''
+    reader = libsbml.SBMLReader()
+    sbml = reader.readSBMLFromString(sbml_file)
+    sbml_model = sbml.getModel()
+    if sbml_model.getNumReactions() > 250:
+        return ('Warning: The model has more than 250 reactions, which may'\
+                ' slow down the computation time significantly. Proceed'\
+                ' with care.')
+    else: return False    
+
+
 def get_modifiers(reaction):
     '''
     get iterator for the reaction modifiers
@@ -581,20 +640,371 @@ def fmin_differential_evolution(f, x0, population_size=100, generations=20000,
     return best_indiv
 
 
-def xml2html(sbml_file):
+def split_sbtabs(sbtab_strings):
     '''
-    generates html view out of xml file
+    Cuts one SBtab string in single SBtab strings if necessary.
+
+    Parameters
+    ----------
+    sbtab_strings: str
+        SBtab table or tables in string representation.
+
+    Returns: list
+        List of SBtab tables in string representation.
     '''
-    old_sbml = str(sbml_file).split('\\n')
-    new_sbml = '<xmp>'
+    sbtabs = []
+    sbtab_string = ''
+    counter = 1
+    
+    for row in sbtab_strings.split('\n'):
+        if row.startswith('!!!') or row.startswith('"!!!'): continue
+        if row.startswith('!!'):
+            if sbtab_string == '':
+                sbtab_string = row + '\n'
+                continue
+            else:
+                try:
+                    if sbtab_string.startswith('!!SBtab') or sbtab_string.startswith('!!ObjTables'):
+                        sbtabs.append(sbtab_string)
+                        counter += 1
+                    sbtab_string = row + '\n'
+                except:
+                    print('Warning: Could not write SBtab %s' % counter)
+                    counter += 1
+        else:
+            sbtab_string += row + '\n'
 
-    for row in old_sbml:
-        new_sbml += row + '\n'
+    if sbtab_string.startswith('!!SBtab') or sbtab_string.startswith('!!ObjTables'):
+        sbtabs.append(sbtab_string)
+                    
+    return sbtabs
 
-    new_sbml += '</xmp>'
+
+def sbtab_to_html(sbtab, filename=None, mode='sbtab_online', template = [], put_links = True, title_string='', show_header_row = True, show_table_name = False, show_table_text = False, definitions_file=''):
+    '''
+    Generates html view out of SBtab table or SBtab document object.
+
+    Parameters
+    ----------
+    sbtab: SBtab.SBtabTable | SBtab.SBtabDocument
+        Either SBtab table object or SBtab document object.
+    filename: str
+        File name of the SBtab table.
+    mode: str
+        Defines the type of HTML to be generated ('sbtab_online' for the SBtab online
+        interface or 'standalone' for a sole HTML page without online binding).
+
+    Returns: str
+        SBtab object as HTML string.
+    '''
+    def _is_float(value):
+        '''
+        checks if an element is a float in string format
+        '''
+        try:
+            float(value)
+            return True
+        except: return False
         
-    return new_sbml
+    def _build_main(sbtab, sbtab_def):
+        '''
+        builds main body of HTML, which needs to be repeated
+        for SBtab Documents
+        '''
+        no_link = ['(',')','+','-','<=>','or','and','FbcOr','FbcAnd']
+        # get column descriptions for this table type and possible shortname links
+        try: (col2description,col2link) = find_descriptions(sbtab_def, sbtab.table_type)
+        except:
+            col2description = False
+            col2link = False
+        
+        # start main
+        html = '<table class="table-striped">'
 
+        # table name
+        if show_table_name:
+            html += '<center><h2>%s</h2></center>' % (sbtab.get_attribute('TableName'))
+
+        if show_table_text:
+            if len(sbtab.get_attribute('Text')):
+                html += '<center><p>%s</p></center>' % (sbtab.get_attribute('Text'))
+
+        # header row
+        #html += '<thead><tr><th colspan="%s">%s</th></tr></thead>' % (len(sbtab.columns), sbtab.header_row)
+        if show_header_row:
+            html += '<h4>%s</h4>' % (sbtab.header_row)
+
+        # columns
+        html += '<thead>'
+        html += '<tr style="line-height:2;">'
+        for col in sbtab.columns:
+            try: title = col2description[col[1:]]
+            except: title = ''
+            html += '<th title="%s">%s</th>' % (title, col)
+        html += '</tr>'
+        html += '</thead>'
+        html += '<tbody>'
+
+        # value rows
+        for row in sbtab.value_rows:
+            # set anchor for internal jump links
+            try: html += '<tr id="%s" style="line-height:1.5;">' % row[sbtab.columns_dict['!ID']]
+            except: html += '<tr style="line-height:1.5;">'
+            for i,col in enumerate(row):
+                # try and set internal jump links via shortnames
+                try:
+                    col2link[sbtab.columns[i]]
+                    if col2link[sbtab.columns[i]] == 'True' and col != '' and col != False:
+                        try:
+                            html += '<td>'
+                            split_column = col.split(' ')
+                            for element in split_column:
+                                if element not in no_link and not _is_float(element) and put_links:
+                                    #html += '<a href="#%s">%s</a> ' % (element, element)    #internal links
+                                    html += element
+                                else:
+                                    html += element + ' '
+                            html += '</td>'
+                        except: html += '<td>%s</td>' % (col)
+                    else:
+                        html += '<td>%s</td>' % (col)
+                except:
+                    if '!Identifiers' in sbtab.columns[i]:
+                        try:
+                            db = re.search('Identifiers:(.*)',sbtab.columns[i])
+                            url = 'http://identifiers.org/%s/%s' % (db.group(1), col)                        
+                            html += '<td><a href="%s">%s</a></td>' % (url, col)
+                        except: html += '<td>%s</td>' % (col)
+                    else:
+                        html += '<td>%s</td>' % (col)
+            html += '</tr>'
+
+        # comment rows
+        for row in sbtab.comments:
+            html += '<tr>'
+            for col in row:
+                html += '<td>%s</td>' % col
+            html += '</tr>'
+
+        # close table
+        html += '</tbody></table>'
+
+        return html
+    
+    ##############################################################################
+    # read in header and footer from HTML template
+    if mode == 'sbtab_online':
+        p = os.path.join(os.path.dirname(__file__), '../modules/template_sbtab_online.html')
+        try:
+            html = open(p, 'r')
+            html_template = html.read()
+            html.close()
+        except:
+            print('HTML template was not found.')
+            return False
+    elif mode == 'standalone':
+        html_template = False
+        try_paths = ['html_templates/template_standalone.html',                     
+                     os.path.join(os.path.dirname(__file__), '../html_templates/template_standalone.html'),
+                     os.path.join(os.path.dirname(__file__), 'html_templates/template_standalone.html'),
+                     template]
+        for path in try_paths:
+            try:
+                html = open(path, 'r')
+                html_template = html.read()
+                html.close()
+            except: pass
+        if not html_template:
+            print('HTML template was not found.')
+            return False
+    else:
+        print('Invalid mode %s. Please use either "sbtab_online" or "standalone".' % mode)
+        return False
+
+    try:
+        header = re.search('(<html lang="en">.*<main>)', html_template, re.DOTALL).group(0)
+        footer = re.search('(</main>.*</html>)', html_template, re.DOTALL).group(0)
+    except:
+        print('Cannot read required template.html.')
+        return False
+
+    html = header
+
+    try:
+        ot = sbtab.object_type
+    except:
+        print('You have not provided a valid SBtab object as input.')
+        return False
+    
+    # replace title placeholder with actual title
+    if len(title_string):
+        html = html.replace('TitlePlaceholder',title_string)
+    else:
+        html = html.replace('TitlePlaceholder',sbtab.filename)
+
+    # read in definitions file for nice mouse over
+    if mode == 'standalone' and len(definitions_file):
+        sbtab_def = open_definitions_file(definitions_file)
+    else:
+        sbtab_def = open_definitions_file()
+        
+    # now build the html file
+    if sbtab.object_type == 'table':
+        html += _build_main(sbtab, sbtab_def)
+    elif sbtab.object_type == 'doc':
+        for sbtab in sbtab.sbtabs:
+            html += _build_main(sbtab, sbtab_def) + '<br><hr>'
+    else:
+        print('The given SBtab object is invalid.')
+        return False
+            
+    html += footer
+    
+    return html
+
+
+def open_definitions_file(_path=None):
+    '''
+    Opens the SBtab definitions file, which can be in several locations.
+
+    Parameters
+    ----------
+    _path: str
+        Optional path to the definitions.tsv.
+
+    Returns: SBtab.SBtabTable
+        SBtab definitions file as SBtabTable object.        
+    '''
+    sbtab_def = False
+    
+    if _path: try_paths = [_path]
+    else:
+        try_paths = ['definitions.tsv',
+                     os.path.join(os.path.dirname(__file__), '../static/files/default_files/definitions.tsv'),
+                     os.path.join(os.path.dirname(__file__), '../definition_table/definitions.tsv'),
+                     os.path.join(os.path.dirname(__file__), 'definitions.tsv'),
+                     os.path.join(os.path.dirname(__file__), 'files/default_files/definitions.tsv')]
+
+    for path in try_paths:
+        try:
+            def_file = open(path, 'r')
+            file_content = def_file.read()
+            sbtab_def = SBtab.SBtabTable(file_content, 'definitions.tsv')
+            def_file.close()
+            break
+        except: pass
+
+    return sbtab_def
+
+
+def extract_supported_table_types():
+    '''
+    Extracts all allowed SBtab table types from the definitions file.
+
+    Returns: list
+        List of supported SBtab table types.
+    '''
+    sbtab_def = open_definitions_file()
+    
+    supported_types = []
+    for row in sbtab_def.value_rows:
+        t = row[sbtab_def.columns_dict['!Parent']]
+        if t not in supported_types and t != 'SBtab':
+            supported_types.append(t)
+
+    return supported_types
+
+
+def find_descriptions(def_file, table_type):
+    '''
+    Preprocesses the definitions file in order to enable some nice mouseover effects for the known column names.
+
+    Parameters
+    ----------
+    def_file: SBtab.SBtabTable
+        Definitions file as SBtab table object.
+    table_type: str
+        SBtab table type for which the descriptions shall be extracted.
+
+    Returns: (dict, dict)
+        Two dictionaries that link the column name to its description and to a Bool flag indicating if it
+        is a shortname identifier possibly linking to another SBtab table.
+    '''
+    col2description = {}
+    col2link = {}
+
+    for row in def_file.value_rows:
+        if row[def_file.columns_dict['!Parent']] == table_type:
+            col2description[row[def_file.columns_dict['!Name']]] = row[def_file.columns_dict['!Description']]
+            col2link['!'+row[def_file.columns_dict['!Name']]] = row[def_file.columns_dict['!isShortname']]
+
+    return (col2description, col2link)
+
+
+def xlsx_to_tsv(file_object, f='web'):
+    '''
+    Converts xlsx SBtab file to tsv format.
+
+    Parameters
+    ----------
+    file_object: xlsx file object
+        SBtab table as xlsx file object.
+    f: str
+        String indicating how the file object is represented ('web' for online file handling, otherwise normal file handling)
+    
+    Returns: str
+        SBtab table as tsv string.
+    '''
+    import openpyxl
+    from io import BytesIO
+
+    if f == 'web': wb = openpyxl.load_workbook(filename = BytesIO(file_object))
+    else: wb = openpyxl.load_workbook(filename = file_object)
+    ws = wb.active
+    ranges = wb[ws.title]
+    table_string = ''
+
+    for row in ranges:
+        for column in row:
+            if str(row[0].value).startswith('!'):
+                if column.value != None and str(column.value) != '':
+                    table_string += str(column.value) + '\t'
+            else:
+                table_string += str(column.value) + '\t'
+        table_string = table_string[:-1] + '\n'
+        
+    return table_string
+
+
+def tab_to_xlsx(sbtab_object):
+    '''
+    Converts SBtab object to xlsx file object.
+
+    Parameters
+    ----------
+    sbtab_object: SBtab.SBtabTable
+        SBtab table object.
+
+    Returns: xlsx file object
+        SBtab table as xlsx file object.
+    '''
+    import openpyxl
+
+    wb = openpyxl.Workbook()
+    ws = wb.active
+    
+    ws['A1'] = sbtab_object.header_row
+    ws.append(sbtab_object.columns)
+    for row in sbtab_object.value_rows:
+        ws.append(row)
+
+    wb.save('transition.xlsx')
+    
+    f = open('transition.xlsx','rb')
+    fileobject = f.read()
+    f.close()
+
+    return fileobject
 
 def id_checker(sbtab, sbml):
     '''
@@ -617,11 +1027,15 @@ def id_checker(sbtab, sbml):
 
     for row in sbtab.value_rows:
         if len(row) < 3: continue
-        try:
-            s_id = sbtab.columns_dict['!Compound:SBML:species:id']
-            r_id = sbtab.columns_dict['!Reaction:SBML:reaction:id']
+        try: s_id = sbtab.columns_dict['!Compound']
         except:
-            sbtabid2sbmlid.append('Error: The SBtab file lacks either of the obligatory columns "'"!Compound:SBML:species:id"'" or "'"!Reaction:SBML:reaction:id"'" to link the parameter entries to the SBML model species.')
+            try: s_id = sbtab.columns_dict['!Compound:SBML:species:id']
+            except: sbtabid2sbmlid.append('Error: The SBtab file lacks the obligatory column "'"!Compound"'"/"'"!Compound:SBML:species:id"'" to link the parameter entries to the SBML model species.')
+
+        try: r_id = sbtab.columns_dict['!Reaction']
+        except:
+            try: r_id = sbtab.columns_dict['!Reaction:SBML:reaction:id']
+            except: sbtabid2sbmlid.append('Error: The SBtab file lacks the obligatory column "'"!Reaction"'"/"'"!Reaction:SBML:reaction:id"'" to link the parameter entries to the SBML model species.')
         try:
             if row[s_id] != '' and row[s_id] not in species_ids_sbml and row[s_id] != 'nan' and row[s_id] != 'None':
                 sbtabid2sbmlid.append('Warning: The SBtab file holds a species ID which does not comply to any species ID in the SBML file: %s'%(row[s_id]))
@@ -634,66 +1048,27 @@ def id_checker(sbtab, sbml):
     return sbtabid2sbmlid
 
 
-def cut_tabs(sbtab_strings):
+def xml_to_html(sbml_file):
     '''
-    cuts an SBtab document in single SBtab files
+    Generates HTML view of XML (SBML) file.
+
+    Parameters
+    ----------
+    sbml_file: str
+        SBML file in string representation.
+
+    Returns: str
+        SBML file as HTML view.
     '''
-    sbtabs = []
-    sbtab_string = ''
-    counter = 1
+    old_sbml = sbml_file.split('\n')
+    new_sbml = '<xmp>'
+    for row in old_sbml:
+        new_sbml += row + '\n'
+    new_sbml += '</xmp>'
 
-    for row in sbtab_strings.split('\n'):
-        if row.startswith('!!'):
-            if sbtab_string == '':
-                sbtab_string = row + '\n'
-                continue
-            else:
-                try:
-                    sbtabs.append(sbtab_string)
-                    sbtab_string = row + '\n'
-                    counter += 1
-                except:
-                    print('Warning: Could not write SBtab %s' % counter)
-                    counter += 1
-        else:
-            sbtab_string += row + '\n'
-    sbtabs.append(sbtab_string) 
-                    
-    return sbtabs
+    return new_sbml
 
 
-    
-    '''
-    # THIS HERE DOESN'T HELP US WITHOUT THE SBTAB DOCUMENT CLASS
-    sbtabs = []
-    sbtab_string = ''
-    counter = 1
-
-    for row in sbtab_strings.split('\n'):
-        print(row)
-        if row.startswith('!!'):
-            if sbtab_string == '':
-                sbtab_string = row + '\n'
-                continue
-            else:
-                try:
-                    sbtab = SBtab.SBtabTable(sbtab_string, 'sbtab_%s.tsv' % counter)
-                    sbtabs.append(sbtab)
-                    sbtab_string = row + '\n'
-                    print('Tis written')
-                    counter += 1
-                except:
-                    print('Warning: Could not write SBtab %s' % counter)
-                    counter += 1
-        else:
-            sbtab_string += row + '\n'
-
-    sbtab = SBtab.SBtabTable(sbtab_string, 'sbtab_%s.tsv' % counter)
-    sbtabs.append(sbtab) 
-                    
-    return sbtabs
-    '''
-
 def tsv_to_html(sbtab, filename=None):
     '''
     generates html view out of tsv file
@@ -736,6 +1111,7 @@ def tsv_to_html(sbtab, filename=None):
 	<!-- simple right aligned list-->
 	<div class="collapse navbar-collapse" id="bs-example-navbar-collapse-1">
 	  <ul class="nav navbar-nav navbar-right">
+            <li> <a href="../../static/css/css_template/gettingstarted.html" title="Getting started">Getting Started</a></li>
 	    <li> <a href="../../default/balancing.html" title="Go to online balancing">Online Balancing</a></li>
 	    <li> <a href="../../static/css/css_template/documentation.html" title="Documentation and Manuals">Documentation</a></li>
 	    <li> <a href="../../static/css/css_template/download.html" title="Installation and Downloads">Download</a></li>
@@ -760,30 +1136,30 @@ def tsv_to_html(sbtab, filename=None):
 
     sbtab_html += '''
     <main>
-    <div class="container-fluid bg-1 text-center" style="padding-top:50px">
-    <div class="row">
+    <div class="container-fluid bg-1c text-center">
+    <div class="row" style="background-color:#9d9d9d;">
     <div class="col-sm-1"></div>
     <div class="col-sm-10">
-    <table class="table-striped" style="font-size:13px">'''
+    <table class="table-striped" style="font-size:13px;background-color:#fff;padding:3px;">'''
         
     first = True
     for i, row in enumerate(ugly_sbtab):
         # declaration of first SBtab in document
         if row.startswith('!!') and first:
-            sbtab_html += '<tr><th colspan="%s">%s</th></tr>' % (len(ugly_sbtab[i+2]), row)
+            sbtab_html += '<tr><th colspan="%s" style="padding:3px;">%s</th></tr>' % (len(ugly_sbtab[i+2]), row)
             first = False
 
         # conclusion of SBtab and beginning of new SBtab (if there are more than one)
         elif row.startswith('!!'):
-            sbtab_html += '</table><br><table class="table-striped" style="font-size:13px">'
-            sbtab_html += '<tr><th colspan="%s">%s</th></tr>' % (len(ugly_sbtab[i+2]), row)
+            sbtab_html += '</table><br><table class="table-striped" style="font-size:13px;background-color:#fff;">'
+            sbtab_html += '<tr><th colspan="%s" style="padding:3px;">%s</th></tr>' % (len(ugly_sbtab[i+2]), row)
 
         # column header row
         elif row.startswith('!'):
             splitrow = row.split(delimiter)
             sbtab_html += '<tr>'
             for col in splitrow:
-                sbtab_html += '<th>%s</th>' % col
+                sbtab_html += '<th style="padding:3px;">%s</th>' % col
             sbtab_html += '</tr>'
 
         # comment row
@@ -799,7 +1175,7 @@ def tsv_to_html(sbtab, filename=None):
             splitrow = row.split(delimiter)
             sbtab_html += '<tr>'
             for col in splitrow:
-                sbtab_html += '<td>%s</td>' % col
+                sbtab_html += '<td style="padding:3px;">%s</td>' % col
             sbtab_html += '</tr>'
 
         '''
